print("import: 'photutils'")
import photutils

print("import: 'photutils.aperture'")
import photutils.aperture

print("import: 'photutils.background'")
import photutils.background

print("import: 'photutils.centroids'")
import photutils.centroids

print("import: 'photutils.datasets'")
import photutils.datasets

print("import: 'photutils.detection'")
import photutils.detection

print("import: 'photutils.extern'")
import photutils.extern

print("import: 'photutils.geometry'")
import photutils.geometry

print("import: 'photutils.isophote'")
import photutils.isophote

print("import: 'photutils.morphology'")
import photutils.morphology

print("import: 'photutils.psf'")
import photutils.psf

print("import: 'photutils.psf.matching'")
import photutils.psf.matching

print("import: 'photutils.segmentation'")
import photutils.segmentation

print("import: 'photutils.utils'")
import photutils.utils

