# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains tools for making or loading datasets for
examples and tests.
"""

from .load import *  # noqa
from .make import *  # noqa
